<?php

/**
 * @file
 * Implements World of Warcraft guild tabard image creation.
 */

/**
 * A World of Warcraft guild tabard renderer.
 *
 * Requires PHP's GD image library.
 */
abstract class BlizzardApiWowTabard {
  /**
   * The generated tabard's maximum size: 240 pixels.
   */
  const TABARD_MAX_SIZE = 240;
  /**
   * The generated tabard's minimum size: 16 pixels.
   */
  const TABARD_MIN_SIZE = 16;
  
  /**
   * Copies a rectangular portion of a tabard base image onto a target image.
   *
   * All coordinates refer to the upper left corner.
   * 
   * @param resource $dst
   *   A GD image handle of the target image.
   * @param string $src
   *   Path of the image to overlay on the target. This must be a PNG image.
   * @param array $dst_point
   *   The position to place the source image on the destination.
   *
   *   Requires the following keys:
   *   - x: The x-coordinate of the destination point.
   *   - y: The y-coordinate of the destination point.
   * @param array $src_rect
   *   (optional) The area of the source image to copy. Defaults to the entire
   *   image.
   *
   *   Requires the following keys:
   *   - x: The x-coordinate of the source point.
   *   - y: The y-coordinate of the source point.
   *   - width: The width of the source area.
   *   - height: The height of the source area.
   * @param mixed $overlay_color
   *   (optional) The color to overlay on the source image.
   */
  protected static function copyResampledImage($dst, $src, $dst_point, $src_rect = array(), $overlay_color = NULL) {
    $src = imagecreatefrompng($src);
    
    // Hexadecimal -> RGB
    if (is_string($overlay_color)) {
      $overlay_color = self::unpackColor($overlay_color);
    }
    // RGB -> GD color
    if (is_array($overlay_color)) {
      $overlay_color = imagecolorallocate($src, $overlay_color['r'], $overlay_color['g'], $overlay_color['b']);
    }
    
    $width = imagesx($src);
    $height = imagesy($src);
    
    if (!is_null($overlay_color)) {
      imagelayereffect($src, IMG_EFFECT_OVERLAY);
      imagefilledrectangle($src, 0, 0, $width, $height, $overlay_color);
    }
    
    // The tabard is always a square, so width and height sizes are the same.
    $size = imagesx($dst);
    // Scale the source image onto the destination image at the given point.
    $dst_rect = array(
      'x' => $size * $dst_point['x'] / self::TABARD_MAX_SIZE,
      'y' => $size * $dst_point['y'] / self::TABARD_MAX_SIZE,
      'width' => $size * $width / self::TABARD_MAX_SIZE,
      'height' => $size * $height / self::TABARD_MAX_SIZE
    );
    // Copy the entire source image if a rectangle wasn't provided.
    $src_rect += array(
      'x' => 0,
      'y' => 0,
      'width' => $width,
      'height' => $height
    );
    
    imagecopyresampled($dst, $src,
      $dst_rect['x'], $dst_rect['y'], $src_rect['x'], $src_rect['y'],
      $dst_rect['width'], $dst_rect['height'], $src_rect['width'], $src_rect['height']);
    imagedestroy($src);
  }
  
  /**
   * Creates a blank PNG image of the specified size for use with GD functions.
   * 
   * Remember to destroy this resource when you're done!
   *
   * @param int $width
   *   The desired image width.
   * @param int $height
   *   The desired image height.
   *
   * @return resource
   *   A GD image handle.
   */
  protected static function createImageResource($width, $height) {
    $res = imagecreatetruecolor($width, $height);
    
    // PNG images only!
    imagealphablending($res, FALSE);
    $transparent = imagecolorallocatealpha($res, 0, 0, 0, 127);
    imagefill($res, 0, 0, $transparent);
    imagealphablending($res, TRUE);
    imagesavealpha($res, TRUE);
    
    return $res;
  }
  
  /**
   * Determines the destination path for a rendered tabard image.
   *
   * Note: You should use file_exists() to determine if the file named in this
   * path actually exists.
   *
   * @param array $guild
   *   (optional) Information on the owner of a tabard. If not provided, the path
   *   of a default tabard image will be returned. Requires the following keys:
   *   - name: The guild name. If transliteration is available, it will be used
   *     to generate a prettier file name.
   *   - realm: The name (or slug) of the realm that the guild is on.
   *     This name will be validated to generate a unique tabard path.
   * @param array $options
   *   (optional) An array with additional options to use when validating the
   *   realm name. Any combination of the following keys are supported:
   *   - region: The Battle.net region to request data from.
   *   - language: The language to request data in.
   *
   * @return mixed
   *   The full path of the tabard image, or FALSE if there was a problem
   *   setting up the directory.
   */
  public static function getDestination($guild = array(), $options = array()) {
    if (empty($guild)) {
      return self::prepareRenderPath('default.png');
    }
    
    $options += array('region' => BlizzardApi::REGION_US, 'language' => BlizzardApi::LANGUAGE_DEFAULT);
    $slug = blizzardapi_realm_slug($guild['realm'], $options, TRUE);
    if (empty($slug)) {
      if ($slug == '') {
        watchdog('blizzardapi', 'Cannot determine tabard destination path: realm not found (%region: %name).',
          array('%region' => $options['region'], '%name' => $guild['realm']), WATCHDOG_DEBUG);
      }
      return FALSE;
    }
    
    // Convert the guild name into a safe file name.
    if (function_exists('transliteration_clean_filename')) {
      $name = transliteration_clean_filename($guild['name'], language_default('language'));
    }
    else {
      $name = drupal_hash_base64($guild['name']);
    }
    
    return self::prepareRenderPath("$slug/$name.png");
  }
  
  /**
   * Prepares the tabard render directory.
   *
   * @param string $suffix
   *   (optional) The desired final path, appended to the default render path.
   *
   * @return mixed
   *   The destination file path, or FALSE if there was a problem setting up the
   *   directory.
   */
  protected static function prepareRenderPath($suffix = '') {
    $destination = file_default_scheme() . '://blizzardapi/tabards/render/' . $suffix;
    $destination = file_stream_wrapper_uri_normalize($destination);
    
    $dirname = drupal_dirname($destination);
    if (!file_prepare_directory($dirname, FILE_CREATE_DIRECTORY)) {
      watchdog('blizzardapi', 'The directory %directory does not exist or is not writable.',
        array('%directory' => $dirname, WATCHDOG_ERROR));
      return FALSE;
    }
    
    return $destination;
  }
  
  /**
   * Renders a World of Warcraft guild tabard image.
   *
   * If the tabard image already exists, it will be overwritten.
   *
   * @param array $tabard
   *   (optional) An array containing tabard creation information (found in
   *   character and guild profiles). If empty, a blank default tabard will be
   *   created.
   *
   *   The following keys are required:
   *   - name: The guild name.
   *   - realm: The name (or slug) of the realm that the guild is on.
   *     This name will be validated to generate a unique tabard path.
   *   - side: The guild faction ID (0: Alliance, 1: Horde).
   *   - emblem: An array containing the following emblem information.
   *     - icon
   *     - iconColor
   *     - border
   *     - borderColor
   *     - backgroundColor
   * @param int $size
   *   (optional) The size of the rendered tabard image. This value must be
   *   within the TABARD_MIN_SIZE and TABARD_MAX_SIZE range.
   * @param array $options
   *   (optional) An array with additional options to use when validating the
   *   realm name. Any combination of the following keys are supported:
   *   - region: The Battle.net region to request data from.
   *   - language: The language to request data in.
   *
   * @return mixed
   *   The path of the generated tabard image, or FALSE if there was a problem
   *   creating the image.
   */
  public static function render($tabard = array(), $size = self::TABARD_MAX_SIZE, $options = array()) {
    if ($size < self::TABARD_MIN_SIZE) {
      $size = self::TABARD_MIN_SIZE;
    }
    elseif ($size > self::TABARD_MAX_SIZE) {
      $size = self::TABARD_MAX_SIZE;
    }
    
    $options += array(
      'region' => BlizzardApi::REGION_US,
      'language' => BlizzardApi::LANGUAGE_DEFAULT
    );
    
    // If Drupal's GD image toolkit is available, then so is the GD library.
    $toolkits = image_get_available_toolkits();
    if (!isset($toolkits['gd'])) {
      drupal_set_message(t('Cannot create tabard image: GD library missing'), 'error');
      return FALSE;
    }
    
    // The following images are added to the blank target image in the following order.
    // Each image supports the following keys:
    // - file: The file name of the image to add.
    // - position: The coordinates on the target image to place this image.
    // - color: (optional) If specified, this color will be blended with the
    //   grey pixels of the source image.
    $images = array(
      'ring' => array(
        'file' => 'ring-alliance.png',
        'position' => array('x' => 0, 'y' => 0)
      ),
      'shadow' => array(
        'file' => 'shadow_00.png',
        'position' => array('x' => 18, 'y' => 27)
      ),
      'background' => array(
        'file' => 'bg_00.png',
        'position' => array('x' => 18, 'y' => 27)
      ),
      'overlay' => array(
        'file' => 'overlay_00.png',
        'position' => array('x' => 18, 'y' => 27)
      ),
      'border' => array(
        'file' => '',
        'position' => array('x' => 31, 'y' => 40)
      ),
      'icon' => array(
        'file' => '',
        'position' => array('x' => 33, 'y' => 57)
      ),
      'hooks' => array(
        'file' => 'hooks.png',
        'position' => array('x' => 18, 'y' => 27)
      )
    );
    
    if (isset($tabard['emblem'])) {
      $images['background']['color'] = $tabard['emblem']['backgroundColor'];
      $images['border']['color'] = $tabard['emblem']['borderColor'];
      $images['icon']['color'] = $tabard['emblem']['iconColor'];
      
      if ($tabard['side'] == 1) {
        $images['ring']['file'] = 'ring-horde.png';
      }
      
      // If the API returns -1 for these values, fall back to the default tabard.
      // Presumably this happens if a guild was just created, but a tabard
      // hasn't been setup yet.
      if ($tabard['emblem']['icon'] < 0 || $tabard['emblem']['border'] < 0) {
        unset($tabard['emblem']);
      }
      else {
        $images['border']['file'] = sprintf('border_%02d.png', $tabard['emblem']['border']);
        $images['icon']['file'] = sprintf('emblem_%02d.png', $tabard['emblem']['icon']);
      }
    }
    
    if (empty($tabard['emblem'])) {
      unset($images['border'], $images['icon']);
    }
    
    $target = self::createImageResource($size, $size);
    
    foreach ($images as $layer => $info) {
      $source = self::retrieveBaseImage($info['file'], $options['region']);
      if ($source == FALSE) {
        imagedestroy($target);
        return FALSE;
      }
      
      $color = isset($info['color']) ? $info['color'] : NULL;
      self::copyResampledImage($target, $source, $info['position'], array(), $color);
    }
    
    $destination = self::getDestination($tabard, $options);
    if (empty($destination)) {
      return FALSE;
    }
    
    $success = self::saveImage($target, $destination);
    imagedestroy($target);
    
    return $success ? $destination : FALSE;
  }
  
  /**
   * Gets the path to a tabard base image. If there is no local copy of the
   * image, it will be downloaded from the specified region's Battle.net site.
   *
   * @param string $filename
   *   The name of the tabard base image.
   * @param string $region
   *   (optional) The Battle.net region to download the image from.
   *   Valid regions include:
   *   - us: Americas (default)
   *   - eu: Europe
   *   - kr: Korea
   *   - tw: Taiwan
   *   - cn: China
   *
   * @return string
   *   The location that the image was saved to, or FALSE if there was a problem
   *   retrieving the file.
   */
  protected static function retrieveBaseImage($filename, $region = BlizzardApi::REGION_US) {
    $destination = file_default_scheme() . '://blizzardapi/tabards/base/' . drupal_basename($filename);
    $destination = file_stream_wrapper_uri_normalize($destination);
    
    $dirname = drupal_dirname($destination);
    if (!file_prepare_directory($dirname, FILE_CREATE_DIRECTORY)) {
      watchdog('blizzardapi', 'The directory %directory does not exist or is not writable.',
        array('%directory' => $dirname, WATCHDOG_ERROR));
      return FALSE;
    }
    
    if (file_exists($destination)) {
      return $destination;
    }
    
    $region = BlizzardApi::isSupportedRegion($region) ? $region : BlizzardApi::REGION_US;
    $host = ($region == BlizzardApi::REGION_CN) ? BlizzardApi::URL_HOST_CHINA : BlizzardApi::URL_HOST;
    $url = 'http:' . $host . BlizzardApiWow::STATIC_IMAGE_PATH . '/guild/tabards/' . $filename;
    $url = str_replace('{region}', $region, $url);
    
    return system_retrieve_file($url, $destination, FALSE, FILE_EXISTS_ERROR);
  }
  
  /**
   * Saves an image to the specified destination (needed for stream wrapper support).
   *
   * This method is a modified version of image_gd_save().
   *
   * @param resource $res
   *   A GD image handle.
   * @param string $destination
   *   A file URI or path where the image should be saved.
   *
   * @return bool
   *   The result of the save operation, either TRUE or FALSE.
   */
  protected static function saveImage($res, $destination) {
    $scheme = file_uri_scheme($destination);
    if ($scheme && file_stream_wrapper_valid_scheme($scheme)) {
      $local_wrappers = file_get_stream_wrappers(STREAM_WRAPPERS_LOCAL);
      if (!isset($local_wrappers[$scheme])) {
        $permanent_destination = $destination;
        $destination = drupal_tempnam('temporary://', 'gd_');
      }
      $destination = drupal_realpath($destination);
    }
    
    imagealphablending($res, FALSE);
    imagesavealpha($res, TRUE);
    $success = imagepng($res, $destination);
    
    if (isset($permanent_destination) && $success) {
      return (bool) file_unmanaged_move($destination, $permanent_destination, FILE_EXISTS_REPLACE);
    }
    return $success;
  }
  
  /**
   * Converts a hexadecimal color string into its aRGB components.
   */
  protected static function unpackColor($hex, $normalize = FALSE) {
    $rgb = array();
    
    if (strlen($hex) == 8) {
      // Split the segments.
      $alpha = substr($hex, 0, 2);
      $hex = substr($hex, 2);
      
      // GD functions expect a 7-bit alpha (0-127).
      $rgb['a'] = abs(hexdec($alpha) - 255) >> 1;
    }
    
    $c = hexdec($hex);
    
    for ($i = 16; $i >= 0; $i -= 8) {
      $out[] = (($c >> $i) & 0xFF) / ($normalize ? 255 : 1);
    }
    
    $rgb += array(
      'r' => $out[0],
      'g' => $out[1],
      'b' => $out[2]
    );
    
    return $rgb;
  }
}
