<?php

/**
 * @file
 * Provides access to World of Warcraft data provided by the Blizzard Community
 * Platform API.
 */

/**
 * Base Blizzard API class for World of Warcraft.
 */
abstract class BlizzardApiWow extends BlizzardApi {
  /**
   * The relative path of static images used on the Battle.net site (this does
   * not include item icons).
   */
  const STATIC_IMAGE_PATH = '/wow/static/images';
  /**
   * The relative path of images rendered using game content on the Battle.net
   * site.
   */
  const STATIC_RENDER_PATH = '/static-render/{region}';
  
  /**
   * Constructs a BlizzardApiWow object.
   *
   * @param string $region
   *   (optional) The Battle.net region to request data from.
   *   Valid regions include:
   *   - us: Americas (default)
   *   - eu: Europe
   *   - kr: Korea
   *   - tw: Taiwan
   *   - cn: China
   * @param string $language
   *   (optional) The language to request data in.
   *   - If not provided, the site default language may be used.
   *   - If the language is not supported, the Battle.net region default will
   *     be used.
   */
  public function __construct($region = BlizzardApi::REGION_US, $language = BlizzardApi::LANGUAGE_DEFAULT) {
    parent::__construct($region, $language);
    $this->setApiPath($this->apiPath . '/wow');
  }
  
  /**
   * Converts gold, silver, and copper into a single integer.
   *
   * @see BlizzardApiWow::unpackCurrency()
   */
  public static function packCurrency($copper, $silver, $gold) {
    return $copper + ($silver * 100) + ($gold * 10000);
  }
  
  /**
   * Converts a price into gold, silver, and copper components.
   *
   * @param int $amount
   *   The item value, such as an auction's bid or buyout price.
   * @param bool $assoc
   *   If TRUE, return the components in an associative array instead of an
   *   object. Defaults to FALSE.
   *
   * @see BlizzardApiWow::packCurrency()
   */
  public static function unpackCurrency($amount, $assoc = FALSE) {
    // There is no integer division in PHP.
    $unpacked = new stdClass();
    $unpacked->copper = (int) ($amount % 100);
    $unpacked->silver = (int) ($amount / 100 % 100);
    $unpacked->gold   = (int) ($amount / 10000);
    
    return $assoc ? (array) $unpacked : $unpacked;
  }
}

/**
 * Provides data about achievements.
 *
 * Default cache lifetime: 30 days.
 */
class BlizzardApiWowAchievement extends BlizzardApiWow {
  /**
   * Local copy of the achievement ID.
   *
   * @var int
   */
  protected $achievementId;
  
  /**
   * Constructs a BlizzardApiWowAchievement object.
   *
   * @param int $achievementId
   *   The ID number of the achievement.
   * @param array $options
   *   (optional) An array containing any combination of the following keys:
   *   - region: The Battle.net region to request data from.
   *   - language: The language to request data in.
   */
  public function __construct($achievementId, $options = array()) {
    $options += array(
      'region' => self::REGION_US,
      'language' => self::LANGUAGE_DEFAULT
    );
    
    parent::__construct($options['region'], $options['language']);
    $this->setApiType('Achievement')
     ->setApiPath($this->apiPath . '/achievement/{achievementid}')
     ->setAchievementId($achievementId);
  }
  
  /**
   * Returns the achievement ID.
   */
  public function getAchievementId() {
    return $this->achievementId;
  }
  
  /**
   * Set the achievement to retrieve.
   *
   * @param int $achievementId
   *   The ID number of the achievement.
   *
   * @throws DomainException, InvalidArgumentException
   */
  public function setAchievementId($achievementId) {
    if (!is_numeric($achievementId)) {
      throw new InvalidArgumentException('Achievement ID must be a number');
    }
    if ($achievementId < 1) {
      throw new DomainException('Achievement ID must be greater than 0');
    }
    
    $this->achievementId = $achievementIdId;
    $this->addPlaceholder('{achievementid}', $achievementId);
    return $this;
  }
}

/**
 * Provides arena ladder information for a specified battlegroup.
 *
 * Developer note: Although the maximum number of results per ladder can never
 * exceed 2000, it can be lower and varies by battlegroup. Requesting more than
 * this number using the page and size methods may result in an empty response.
 *
 * Default cache lifetime: 1 day.
 *
 * @see BlizzardApiWowRatedBattlegroundLadder
 */
class BlizzardApiWowArenaLadder extends BlizzardApiWow {
  const TEAMSIZE_2v2 = '2v2';
  const TEAMSIZE_3v3 = '3v3';
  const TEAMSIZE_5v5 = '5v5';
  
  private static $schema = array(
    'teamsize' => array(
      self::TEAMSIZE_2v2,
      self::TEAMSIZE_3v3,
      self::TEAMSIZE_5v5
    )
  );
  
  /**
   * Local copy of the specified battlegroup.
   * 
   * @var string
   */
  protected $battlegroup;
  
  /**
   * Local copy of the ladder sorting method.
   * 
   * @var bool
   */
  protected $sort;
  
  /**
   * Local copy of the ladder team size to get.
   * 
   * @var string
   */
  protected $teamsize;
  
  /**
   * Constructs a BlizzardApiWowArenaLadder object.
   *
   * @param string $battlegroup
   *   A realm battlegroup.
   * @param string $teamsize
   *   An arena team size, supported values include: 2v2, 3v3, and 5v5.
   * @param int $size
   *   (optional) The number of entries per page. Defaults to 50.
   * @param int $page
   *   (optional) The page number to retrieve. Defaults to 1.
   * @param bool $asc
   *   (optional) If TRUE, sort the ladder in ascending order. Defaults to TRUE.
   * @param array $options
   *   (optional) An array containing any combination of the following keys:
   *   - region: The Battle.net region to request data from.
   *   - language: The language to request data in.
   */
  public function __construct($battlegroup, $teamsize, $size = 50, $page = 1, $asc = TRUE, $options = array()) {
    $options += array(
      'region' => self::REGION_US,
      'language' => self::LANGUAGE_DEFAULT
    );
    
    parent::__construct($options['region'], $options['language']);
    $this->setApiType('Arena ladder')
      ->setApiPath($this->apiPath . '/pvp/arena/{battlegroup}/{teamsize}')
      ->setCacheLifetime(86400)  // 1 day
      ->setBattlegroup($battlegroup)
      ->setTeamSize($teamsize)
      ->setSize($size)->setPage($page)->setSortOrder($asc);
  }
  
  /**
   * Returns the realm battlegroup of the arena ladder.
   */
  public function getBattlegroup() {
    return $this->battlegroup;
  }
  
  /**
   * Returns the page number to retrieve.
   *
   * @see BlizzardApiWowArenaLadder::getSize()
   */
  public function getPage() {
    return $this->fields['page'];
  }
  
  /**
   * Returns the number of entries allowed per page.
   *
   * @see BlizzardApiWowArenaLadder::getPage()
   */
  public function getSize() {
    return $this->fields['size'];
  }
  
  /**
   * Returns TRUE if the arena ladder is sorted in ascending order, FALSE otherwise.
   */
  public function getSortOrder() {
    return $this->sort;
  }
  
  /**
   * Returns a list of supported arena team sizes.
   */
  public static function getSupportedTeamSizes() {
    return self::$schema['teamsize'];
  }
  
  /**
   * Returns the team size of the arena ladder.
   */
  public function getTeamSize() {
    return $this->teamsize;
  }
  
  /**
   * Overrides BlizzardApi::prepareData().
   */
  public function prepareData() {
    $this->data = $this->data['arenateam'];
  }
  
  /**
   * Set realm battlegroup of the arena ladder.
   *
   * @param string $battlegroup
   *   A realm battlegroup.
   *
   * @throws InvalidArgumentException
   */
  public function setBattlegroup($battlegroup) {
    if (empty($battlegroup)) {
      throw new InvalidArgumentException('Battlegroup cannot be empty');
    }
    $this->battlegroup = $battlegroup;
    $this->addPlaceholder('{battlegroup}', $battlegroup);
    return $this;
  }
  
  /**
   * Set the arena ladder page to retrieve.
   *
   * @param int $page
   *   (optional) The ladder page to retrieve. Defaults to 1.
   *
   * @see BlizzardApiWowArenaLadder::setSize()
   *
   * @throws DomainException, InvalidArgumentException
   */
  public function setPage($page = 1) {
    if (!is_numeric($page)) {
      throw new InvalidArgumentException('Page must be a number');
    }
    if ($page < 1) {
      throw new DomainException('Page cannot be less than 1');
    }
    $this->fields['page'] = $page;
    return $this;
  }
  
  /**
   * Set the number of results to return per arena ladder page.
   *
   * @param int $size
   *   (optional) The number of entries per page. Defaults to 50.
   *
   * @see BlizzardApiWowArenaLadder::setPage()
   *
   * @throws DomainException, InvalidArgumentException
   */
  public function setSize($size = 50) {
    if (!is_numeric($size)) {
      throw new InvalidArgumentException('Size must be a number');
    }
    if ($size < 1 || $size > 2000) {
      throw new DomainException('Size must be within the range 1-2000');
    }
    
    $this->fields['size'] = $size;
    return $this;
  }
  
  /**
   * Set the order of the retrieved arena ladder.
   *
   * @param bool $asc
   *   (optional) If TRUE, the ladder will be sorted in ascending order.
   *   Defaults to TRUE.
   */
  public function setSortOrder($asc = TRUE) {
    $this->sort = (bool) $asc;
    $this->fields['asc'] = $asc ? 'true' : 'false';
    return $this;
  }
  
  /**
   * Set the arena team's size.
   *
   * @param string $teamSize
   *   An arena team size, supported values include: 2v2, 3v3, and 5v5.
   *
   * @throws InvalidArgumentException
   */
  public function setTeamSize($teamSize) {
    if (!in_array($teamSize, self::$schema['teamsize'])) {
      throw new InvalidArgumentException("Invalid team size: $teamSize");
    }
    
    $this->teamsize = $teamSize;
    $this->addPlaceholder('{teamsize}', $teamSize);
    return $this;
  }
}

/**
 * Provides detailed arena team information.
 *
 * Default cache lifetime: 1 day.
 */
class BlizzardApiWowArenaTeam extends BlizzardApiWow {
  const TEAMSIZE_2v2 = '2v2';
  const TEAMSIZE_3v3 = '3v3';
  const TEAMSIZE_5v5 = '5v5';
  
  private static $schema = array(
    'teamsize' => array(
      self::TEAMSIZE_2v2,
      self::TEAMSIZE_3v3,
      self::TEAMSIZE_5v5
    )
  );
  
  /**
   * Local copy of the arena team's realm name.
   *
   * @var string
   */
  protected $realm;
  
  /**
   * Local copy of the arena team's size.
   *
   * @var string
   */
  protected $teamSize;
  
  /**
   * Local copy of the arena team's name.
   *
   * @var string
   */
  protected $teamName;
  
  /**
   * Constructs a BlizzardApiWowArenaTeam object.
   *
   * @param string $realm
   *   The realm that the arena team is on. This should preferably be the realm
   *   slug.
   * @param string $teamSize
   *   An arena team size, supported values include: 2v2, 3v3, and 5v5.
   * @param string $teamName
   *   The name of the arena team to get.
   * @param array $options
   *   (optional) An array containing any combination of the following keys:
   *   - region: The Battle.net region to request data from.
   *   - language: The language to request data in.
   */
  public function __construct($realm, $teamSize, $teamName, $options = array()) {
    $options += array(
      'region' => BlizzardApi::REGION_US,
      'language' => BlizzardApi::LANGUAGE_DEFAULT
    );
    
    parent::__construct($options['region'], $options['language']);
    $this->setApiType('Arena team')
      ->setApiPath($this->apiPath . '/arena/{realm}/{teamsize}/{teamname}')
      ->setCacheLifetime(86400)  // 1 day
      ->setRealm($realm)
      ->setTeamSize($teamSize)
      ->setTeamName($teamName);
  }
  
  /**
   * Returns the name of the realm that arena team data was requested for.
   */
  public function getRealmName() {
    return $this->realm;
  }
  
  /**
   * Returns a list of supported arena team sizes.
   */
  public static function getSupportedTeamSizes() {
    return self::$schema['teamsize'];
  }
  
  /**
   * Returns the requested arena team's name.
   */
  public function getTeamName() {
    return $this->teamName;
  }
  
  /**
   * Returns the requested arena team's size.
   */
  public function getTeamSize() {
    return $this->teamSize;
  }
  
  /**
   * Set the realm to request arena team data from.
   *
   * @param string $realm
   *   The name of the realm. This should preferably be the realm slug.
   *
   * @throws InvalidArgumentException
   */
  public function setRealm($realm) {
    if (empty($realm)) {
      throw new InvalidArgumentException('Realm cannot be empty');
    }
    
    $this->realm = $realm;
    $this->addPlaceholder('{realm}', $realm);
    return $this;
  }
  
  /**
   * Set the arena team's name.
   *
   * @param string $teamName
   *   The name of the arena team to get.
   *
   * @throws InvalidArgumentException
   */
  public function setTeamName($teamName) {
    if (empty($teamName)) {
      throw new InvalidArgumentException('Team name cannot be empty');
    }
    
    $this->teamName = $teamName;
    $this->addPlaceholder('{teamname}', $teamName);
    return $this;
  }
  
  /**
   * Set the arena team's size.
   *
   * @param string $teamSize
   *   An arena team size, supported values include: 2v2, 3v3, and 5v5.
   *
   * @throws InvalidArgumentException
   */
  public function setTeamSize($teamSize) {
    if (!in_array($teamSize, self::$schema['teamsize'])) {
      throw new InvalidArgumentException("Invalid team size: $teamSize");
    }
    
    $this->teamsize = $teamSize;
    $this->addPlaceholder('{teamsize}', $teamSize);
    return $this;
  }
}

/**
 * Provides information on the most recent batch of current auction data.
 *
 * Default cache lifetime: 1 hour.
 */
class BlizzardApiWowAuction extends BlizzardApiWow {
  /**
   * Less than 30 minutes.
   */
  const TIME_LEFT_SHORT = 'SHORT';
  /**
   * Less than 2 hours.
   */
  const TIME_LEFT_MEDIUM = 'MEDIUM';
  /**
   * Less than 12 hours.
   */
  const TIME_LEFT_LONG = 'LONG';
  /**
   * Greater than 12 hours.
   */
  const TIME_LEFT_VERY_LONG = 'VERY_LONG';
  
  /**
   * Local copy of the specified realm.
   *
   * @var string
   */
  protected $realm = '';
  
  /**
   * Constructs a BlizzardApiWowAuction object.
   *
   * @param string $realm
   *   The realm of the auction house to get data from. This should be a realm
   *   slug, but can also be the display name.
   * @param array $options
   *   (optional) An array containing any combination of the following keys:
   *   - region: The Battle.net region to request data from.
   *   - language: The language to request data in.
   */
  public function __construct($realm, $options = array()) {
    $options += array(
      'region' => BlizzardApi::REGION_US,
      'language' => BlizzardApi::LANGUAGE_DEFAULT
    );
    
    parent::__construct($options['region'], $options['language']);
    $this->setApiType('Auction')
      ->setApiPath($this->apiPath . '/auction/data/{realm}')
      ->setCacheLifetime(3600)  // 1 hour
      ->setRealm($realm);
  }
  
  /**
   * Deletes all downloaded auction files.
   *
   * Warning: Do NOT use this method on managed files!
   *
   * @param string $path
   *   The base directory or URI to scan, without the trailing slash.
   *
   *   Defaults to 'public://blizzardapi/auctions/{region}-{realm}'.
   */
  public function clearAuctionFiles($path = '') {
    $files = $this->getAuctionFiles($path);
    foreach ($files as $timestamp => $info) {
      file_unmanaged_delete($info->uri);
    }
    
    return $this;
  }
  
  /**
   * Downloads the auction dump file returned by this API request to the
   * specified destination.
   *
   * @param string $destination
   *   Stream wrapper URI specifying where the file should be placed.
   * @param bool $managed
   *   If TRUE, the file API hooks will be invoked and the file will be
   *   registered in the database. Defaults to FALSE.
   * @param int $replace
   *   Replace behavior when the destination file already exists:
   *   - FILE_EXISTS_REPLACE: Replace the existing file (default).
   *   - FILE_EXISTS_RENAME: Append _{incrementing number} until the filename is
   *     unique.
   *   - FILE_EXISTS_ERROR: Do nothing and return FALSE.
   * @return mixed
   *   The location that the file was saved to or FALSE if there was an error.
   *
   * @throws BlizzardApiInvalidOperationException
   *
   * @see BlizzardApiWowAuction::prepareDefaultPath()
   */
  public function downloadAuctionFile($destination, $managed = FALSE, $replace = FILE_EXISTS_REPLACE) {
    if (!isset($this->data)) {
      throw new BlizzardApiInvalidOperationException('Cannot download auction dump file until realm index has been retrieved.');
    }
    
    // Blizzard has stopped serving .gz files, but requesting compressed data
    // still saves bandwidth (usually going from ~3MB to ~500KB per file).
    return blizzardapi_retrieve_file($this->data['url'], $destination, $managed, $replace);
  }
  
  /**
   * Finds all downloaded auction files in a given directory.
   *
   * @param string $path
   *   The base directory or URI to scan, without the trailing slash.
   *
   *   Defaults to 'public://blizzardapi/auctions/{region}-{realm}'.
   * @param string $key
   *   The key used for the returned associative array of files. Defaults to
   *   'name'.
   *
   *   Possible keys include:
   *   - filename: The file's name.
   *   - name: The file's name, without the extension.
   *   - uri: The file's URI.
   *
   * @return mixed
   *   A list of auction files, or FALSE if the default path could not be determined.
   */
  public function getAuctionFiles($path = '', $key = 'name') {
    if (empty($path)) {
      $options = array('region' => $this->region, 'language' => $this->language);
      $slug = blizzardapi_realm_slug($this->realm, $options, TRUE);
      if (empty($slug)) {
        return FALSE;
      }
      
      $path = file_default_scheme() . '://blizzardapi/auctions/' . $slug;
    }
    
    // This method is also called when deleting files, so be picky about filenames.
    return file_scan_directory($path, '/[0-9]{13}\.json$/', array('key' => $key));
  }
  
  /**
   * Returns the name of the realm that auction data was requested for.
   */
  public function getRealmName() {
    return $this->realm;
  }
  
  /**
   * Overrides BlizzardApi::prepareData().
   */
  protected function prepareData() {
    // Let's see if Blizzard ever includes more than 1 file...
    if (count($this->data['files']) > 1) {
      watchdog('blizzardapi', 'Multiple auction files found in response from %url.',
        array('%url' => $this->url), WATCHDOG_WARNING);
    }
    $this->data = $this->data['files'][0];
  }
  
  /**
   * Generates an auction dump filename and prepares the default directory to
   * save it in.
   *
   * Example: public://blizzardapi/auctions/{region}-{realm}/{timestamp}.json
   *
   * @return mixed
   *   A destination filename or FALSE if there was a problem setting up the
   *   directory.
   *
   * @see BlizzardApiWowAuction::downloadAuctionFile()
   */
  public function prepareDefaultPath() {
    if (!isset($this->data)) {
      throw new BlizzardApiInvalidOperationException('Cannot prepare auction dump path until realm index has been retrieved.');
    }
    
    $options = array('region' => $this->region, 'language' => $this->language);
    $slug = blizzardapi_realm_slug($this->realm, $options, TRUE);
    if (empty($slug)) {
      return FALSE;
    }
    
    $path = $slug . '/' . $this->data['lastModified'];
    $destination = file_default_scheme() . '://blizzardapi/auctions/' . $path . '.json';
    $destination = file_stream_wrapper_uri_normalize($destination);
    
    $dirname = drupal_dirname($destination);
    if (!file_prepare_directory($dirname, FILE_CREATE_DIRECTORY)) {
      watchdog('blizzardapi', 'The directory %directory does not exist or is not writable.',
        array('%directory' => $dirname, WATCHDOG_ERROR));
      return FALSE;
    }
    
    return $destination;
  }
  
  /**
   * Set the realm to request auction house data from.
   *
   * @param string $realm
   *   The name of the realm. This should preferably be the realm slug.
   *
   * @throws InvalidArgumentException
   */
  public function setRealm($realm) {
    if (empty($realm)) {
      throw new InvalidArgumentException('Realm cannot be empty');
    }
    
    $this->realm = $realm;
    $this->addPlaceholder('{realm}', $realm);
    return $this;
  }
}

/**
 * Provides character profile information.
 *
 * Default cache lifetime: 1 day.
 */
class BlizzardApiWowCharacter extends BlizzardApiWow {
  const FEED_TYPE_ACHIEVEMENT = 'ACHIEVEMENT';
  const FEED_TYPE_BOSSKILL = 'BOSSKILL';
  const FEED_TYPE_CRITERIA = 'CRITERIA';
  const FEED_TYPE_LOOT = 'LOOT';
  
  const GENDER_MALE = 0;
  const GENDER_FEMALE = 1;
  
  const REPUTATION_HATED = 0;
  const REPUTATION_HOSTILE = 1;
  const REPUTATION_UNFRIENDLY = 2;
  const REPUTATION_NEUTRAL = 3;
  const REPUTATION_FRIENDLY = 4;
  const REPUTATION_HONORED = 5;
  const REPUTATION_REVERED = 6;
  const REPUTATION_EXALTED = 7;
  
  private static $schema = array(
    'fields' => array(
      'achievements',
      'appearance',
      'companions',
      'feed',
      'guild',
      'items',
      'mounts',
      'pets',
      'professions',
      'progression',
      'pvp',
      'quests',
      'reputation',
      'stats',
      'talents',
      'titles',
    )
  );
  
  /**
   * Local copy of the specified character.
   *
   * @var string
   */
  protected $character = '';
  
  /**
   * Local copy of the specified realm.
   *
   * @var string
   */
  protected $realm = '';
  
  /**
   * Constructs a BlizzardApiWowCharacter object.
   *
   * @param string $character
   *   The name of the character.
   * @param string $realm
   *   The realm that the character is on. This should be a realm slug, but can
   *   also be the realm's display name.
   * @param array $fields
   *   (optional) A list of additional fields to request.
   * @param array $options
   *   (optional) An array containing any combination of the following keys:
   *   - region: The Battle.net region to request data from.
   *   - language: The language to request data in.
   *
   * @see BlizzardApiWowCharacter::getSupportedFields()
   */
  public function __construct($character, $realm, $fields = array(), $options = array()) {
    $options += array(
      'region' => self::REGION_US,
      'language' => self::LANGUAGE_DEFAULT
    );
    
    parent::__construct($options['region'], $options['language']);
    $this->setApiType('Character')
      ->setApiPath($this->apiPath . '/character/{realm}/{character}')
      ->setCacheLifetime(86400)  // 1 day
      ->setRealm($realm)
      ->setCharacter($character)
      ->setFields($fields);
  }
  
  /**
   * Downloads a profile image of the character.
   *
   * @param bool $refresh
   *   (optional) If TRUE, download the image even if a local copy already
   *   exists. The previous file will be overwritten.
   *
   * @return mixed
   *   The location of the saved image or FALSE if there was an error saving it.
   */
  public function downloadProfileImage($refresh = FALSE) {
    $url = str_replace('-avatar.jpg', '-profilemain.jpg', $this->getThumbnailUrl());
    
    $profile = $this->retrieveImage($url, $refresh);
    if ($profile === FALSE) {
      watchdog('blizzardapi', 'Unable to save image from %remote.', array('%remote' => $url), WATCHDOG_ERROR);
    }
    
    return $profile;
  }
  
  /**
   * Downloads a thumbnail (avatar) image of the character.
   *
   * @param bool $refresh
   *   (optional) If TRUE, download the image even if a local copy already
   *   exists. The previous file will be overwritten.
   *
   * @return mixed
   *   The location of the saved image or FALSE if there was an error saving it.
   */
  public function downloadThumbnailImage($refresh = FALSE) {
    $url = $this->getThumbnailUrl();
    
    $avatar = $this->retrieveImage($url, $refresh);
    if ($avatar === FALSE) {
      watchdog('blizzardapi', 'Unable to save image from %remote.', array('%remote' => $url), WATCHDOG_ERROR);
    }
    
    return $avatar;
  }
  
  /**
   * Returns the name of the character.
   */
  public function getCharacterName() {
    return $this->character;
  }
  
  /**
   * Returns the name of the realm that the character is on.
   */
  public function getRealmName() {
    return $this->realm;
  }
  
  /**
   * Returns a list of supported character fields.
   */
  public static function getSupportedFields() {
    return self::$schema['fields'];
  }
  
  /**
   * Returns the full thumbnail URL.
   *
   * @throws BlizzardApiInvalidOperationException
   */
  public function getThumbnailUrl() {
    if (!isset($this->data['thumbnail'])) {
      throw new BlizzardApiInvalidOperationException('Cannot get URL until character data has been retrieved.');
    }
    
    $url = 'http:' . $this->getApiHost() . BlizzardApiWow::STATIC_RENDER_PATH . '/' . $this->data['thumbnail'];
    return str_replace('{region}', $this->region, $url);
  }
  
  /**
   * Retrieves an image using drupal_http_request() and saves it to a local
   * directory.
   *
   * @param string $source
   *   The URL of the file to get.
   * @param bool $refresh
   *   (optional) If TRUE, download the image even if a local copy already
   *   exists. The previous file will be overwritten.
   *
   * @return mixed
   *   The location of the saved image or FALSE if there was an error saving it.
   */
  protected function retrieveImage($source, $refresh = FALSE) {
    if (!isset($this->data['thumbnail'])) {
      throw new BlizzardApiInvalidOperationException('Cannot determine destination path until character data has been retrieved.');
    }
    
    // Notes for these variables:
    // - Realm will always contain an english slug. PHP shouldn't crash because
    //   of UTF-8 or special characters.
    // - Unknown is an integer that may or may not be required to save unique
    //   character images. Therefore it is included in the URI just in case.
    // - Filename is unused.
    list($realm, $unknown, $filename) = explode('/', $this->data['thumbnail'], 3);
    $destination = file_default_scheme() . '://' . "blizzardapi/characters/$this->region-$realm/$unknown/" . drupal_basename($source);
    $destination = file_stream_wrapper_uri_normalize($destination);
    
    $dirname = drupal_dirname($destination);
    if (!file_prepare_directory($dirname, FILE_CREATE_DIRECTORY)) {
      watchdog('blizzardapi', 'The directory %directory does not exist or is not writable.',
        array('%directory' => $dirname, WATCHDOG_ERROR));
      return FALSE;
    }
    
    if (file_exists($destination) && !$refresh) {
      return $destination;
    }
    
    return system_retrieve_file($source, $destination, FALSE, FILE_EXISTS_REPLACE);
  }
  
  /**
   * Set the character's name.
   *
   * @param string $character
   *
   * @throws InvalidArgumentException
   */
  public function setCharacter($character) {
    if (empty($character)) {
      throw new InvalidArgumentException('Character cannot be empty');
    }
    
    $this->character = $character;
    $this->addPlaceholder('{character}', $character);
    return $this;
  }
  
  /**
   * Set the additional character fields to retrieve.
   *
   * @param array $fields
   *   A list of datasets to get. By default, the character's realm, name, level,
   *   last modified timestamp, thumbnail, race ID, achievement points, gender
   *   ID, and class ID are returned. Any unsupported fields will not be included
   *   in the request.
   *
   * @see BlizzardApiWowCharacter::getSupportedFields()
   */
  public function setFields($fields) {
    $this->fields['fields'] = array_intersect($fields, self::$schema['fields']);
    return $this;
  }
  
  /**
   * Set the character's realm.
   *
   * @param string $realm
   *   The name of the realm. This should preferably be the realm slug.
   *
   * @throws InvalidArgumentException
   */
  public function setRealm($realm) {
    if (empty($realm)) {
      throw new InvalidArgumentException('Realm cannot be empty');
    }
    
    $this->realm = $realm;
    $this->addPlaceholder('{realm}', $realm);
    return $this;
  }
}

/**
 * Provides guild profile information.
 *
 * Default cache lifetime: 1 day.
 */
class BlizzardApiWowGuild extends BlizzardApiWow {
  const NEWS_GUILD_ACHIEVEMENT = 'guildAchievement';
  const NEWS_GUILD_CREATED = 'guildCreated';
  const NEWS_GUILD_LEVEL = 'guildLevel';
  const NEWS_ITEM_LOOT = 'itemLoot';
  const NEWS_ITEM_PURCHASE = 'itemPurchase';
  const NEWS_PLAYER_ACHIEVEMENT = 'playerAchievement';
  
  private static $schema = array(
    'fields' => array(
      'achievements',
      'members',
      'news'
    )
  );
  
  /**
   * Local copy of the specified guild.
   *
   * @var string
   */
  protected $guild = '';
  
  /**
   * Local copy of the specified realm.
   *
   * @var string
   */
  protected $realm = '';
  
  /**
   * Constructs a BlizzardApiWowGuild object.
   *
   * @param string $guild
   *   The name of the guild.
   * @param string $realm
   *   The realm the guild is on. This should be a realm slug, but can also be
   *   the realm's display name.
   * @param array $fields
   *   (optional) A list of additional fields to request.
   * @param array $options
   *   (optional) An array containing any combination of the following keys:
   *   - region: The Battle.net region to request data from.
   *   - language: The language to request data in.
   *
   * @see BlizzardApiWowGuild::getSupportedFields()
   */
  public function __construct($guild, $realm, $fields = array(), $options = array()) {
    $options += array(
      'region' => self::REGION_US,
      'language' => self::LANGUAGE_DEFAULT
    );
    
    parent::__construct($options['region'], $options['language']);
    $this->setApiType('Guild')
      ->setApiPath($this->apiPath . '/guild/{realm}/{guild}')
      ->setCacheLifetime(86400)  // 1 day
      ->setRealm($realm)
      ->setGuild($guild)
      ->setFields($fields);
  }
  
  /**
   * Returns the name of the guild.
   */
  public function getGuildName() {
    return $this->guild;
  }
  
  /**
   * Returns the name of the realm that the guild is on.
   */
  public function getRealmName() {
    return $this->realm;
  }
  
  /**
   * Returns a list of supported guild fields.
   */
  public static function getSupportedFields() {
    return self::$schema['fields'];
  }
  
  /**
   * Set the additional guild fields to retrieve.
   *
   * @param array $fields
   *   A list of datasets to get. By default, the guild's name, level,
   *   faction and achievement points are automatically returned. Any unsupported
   *   fields will not be included in the request.
   *
   * @see BlizzardApiWowGuild::getSupportedFields()
   */
  public function setFields($fields) {
    $this->fields['fields'] = array_intersect($fields, self::$schema['fields']);
    return $this;
  }
  
  /**
   * Set the guild's name.
   *
   * @param string $guild
   *   The name of the guild.
   *
   * @throws InvalidArgumentException
   */
  public function setGuild($guild) {
    if (empty($guild)) {
      throw new InvalidArgumentException('Guild cannot be empty');
    }
    
    $this->guild = $guild;
    $this->addPlaceholder('{guild}', $guild);
    return $this;
  }
  
  /**
   * Set the guild's realm.
   *
   * @param string $realm
   *   The name of the realm. This should preferably be the realm slug.
   *
   * @throws InvalidArgumentException
   */
  public function setRealm($realm) {
    if (empty($realm)) {
      throw new InvalidArgumentException('Realm cannot be empty');
    }
    
    $this->realm = $realm;
    $this->addPlaceholder('{realm}', $realm);
    return $this;
  }
}

/**
 * Provides item information.
 *
 * Default cache lifetime: 30 days.
 *
 * @see BlizzardApiWowItemSet
 */
class BlizzardApiWowItem extends BlizzardApiWow {
  /**
   * Also includes items that bind to a Battle.net account.
   */
  const BINDS_ON_PICKUP = 1;
  const BINDS_ON_EQUIP  = 2;
  const BINDS_ON_USE    = 3;
  
  const INVENTORY_TYPE_NONE     = 0;
  const INVENTORY_TYPE_HEAD     = 1;
  const INVENTORY_TYPE_NECK     = 2;
  const INVENTORY_TYPE_SHOULDER = 3;
  const INVENTORY_TYPE_SHIRT    = 4;
  const INVENTORY_TYPE_CHEST    = 5;
  const INVENTORY_TYPE_WAIST    = 6;
  const INVENTORY_TYPE_LEGS     = 7;
  const INVENTORY_TYPE_FEET     = 8;
  const INVENTORY_TYPE_WRIST    = 9;
  const INVENTORY_TYPE_HAND     = 10;
  const INVENTORY_TYPE_RING     = 11;
  const INVENTORY_TYPE_TRINKET  = 12;
  const INVENTORY_TYPE_ONEHAND  = 13;
  const INVENTORY_TYPE_SHIELD   = 14;
  const INVENTORY_TYPE_BOW      = 15;
  const INVENTORY_TYPE_CLOAK    = 16;
  /**
   * Includes fishing poles.
   */
  const INVENTORY_TYPE_TWOHAND  = 17;
  const INVENTORY_TYPE_BAG      = 18;
  const INVENTORY_TYPE_TABARD   = 19;
  const INVENTORY_TYPE_ROBE     = 20; // Examples: 44692, 34085, 34685.
  const INVENTORY_TYPE_MAINHAND = 21;
  const INVENTORY_TYPE_OFFHAND  = 22;
  const INVENTORY_TYPE_HELD_IN_OFFHAND = 23;
  const INVENTORY_TYPE_AMMO     = 24; // Unused.
  const INVENTORY_TYPE_THROWN   = 25;
  /**
   * Includes guns, crossbows, and wands.
   */
  const INVENTORY_TYPE_RANGED   = 26;
//const INVENTORY_TYPE_UNKNOWN  = 27;
  const INVENTORY_TYPE_RELIC    = 28;
  
  const QUALITY_POOR      = 0;
  const QUALITY_COMMON    = 1;
  const QUALITY_UNCOMMON  = 2;
  const QUALITY_SUPERIOR  = 3;
  const QUALITY_EPIC      = 4;
  const QUALITY_LEGENDARY = 5;
  const QUALITY_ARTIFACT  = 6; // Unused.
  const QUALITY_HEIRLOOM  = 7;
  
  const REFORGE_SPIRIT    = 0;
  const REFORGE_DODGE     = 1;
  const REFORGE_PARRY     = 2;
  const REFORGE_HIT       = 3;
  const REFORGE_CRIT      = 4;
  const REFORGE_HASTE     = 5;
  const REFORGE_EXPERTISE = 6;
  const REFORGE_MASTERY   = 7;
  
  // @todo There may be more unknown item source types. Testing IDs follow each constant.
  const SOURCE_ACHIEVEMENT_REWARD  = 'ACHIEVEMENT_REWARD';  // 71033 Lil' Tarecgosa
  const SOURCE_CREATED_BY_SPELL    = 'CREATED_BY_SPELL';    // 69945 Don Tayo's Inferno Mittens
  const SOURCE_CREATURE_DROP       = 'CREATURE_DROP';       // 56431 Right Eye of Rajh
  const SOURCE_CREATURE_PICKPOCKET = 'CREATURE_PICKPOCKET'; // 38268 Spare Hand
  const SOURCE_FACTION_REWARD      = 'FACTION_REWARD';      // 62366 Arcanum of the Earthen Ring
  const SOURCE_GAME_OBJECT_DROP    = 'GAME_OBJECT_DROP';    // 56181 Alliance Survival Kit
  const SOURCE_NONE                = 'NONE';                // 65499 Foam Sword
  const SOURCE_PROMOTION           = 'PROMOTION';           // 49284 Reins of the Swift Spectral Tiger
  const SOURCE_REWARD_FOR_QUEST    = 'REWARD_FOR_QUEST';    // 49623 Shadowmorne
  const SOURCE_VENDOR              = 'VENDOR';              // 78694 Deep Earth Headpiece
  
  const STAT_HEALTH      = 1;
  const STAT_AGILITY     = 3;
  const STAT_STRENGTH    = 4;
  const STAT_INTELLECT   = 5;
  const STAT_SPIRIT      = 6;
  const STAT_STAMINA     = 7;
  const STAT_DEFENSE     = 12;
  const STAT_DODGE       = 13;
  const STAT_PARRY       = 14;
  const STAT_BLOCK       = 15;
  const STAT_HIT_MELEE   = 16;
  const STAT_HIT_RANGED  = 17;
  const STAT_HIT_SPELL   = 18;
  const STAT_CRIT_MELEE  = 19;
  const STAT_CRIT_RANGED = 20;
  const STAT_CRIT_SPELL  = 21;
  const STAT_HIT_MELEE_TAKEN   = 22;
  const STAT_HIT_RANGED_TAKEN  = 23;
  const STAT_HIT_SPELL_TAKEN   = 24;
  const STAT_CRIT_MELEE_TAKEN  = 25;
  const STAT_CRIT_RANGED_TAKEN = 26;
  const STAT_CRIT_SPELL_TAKEN  = 27;
  const STAT_HASTE_MELEE  = 28;
  const STAT_HASTE_RANGED = 29;
  const STAT_HASTE_SPELL  = 30;
  const STAT_HIT          = 31;
  const STAT_CRIT         = 32;
  const STAT_HIT_TAKEN    = 33;
  const STAT_CRIT_TAKEN   = 34;
  const STAT_RESILLIENCE  = 35;
  const STAT_HASTE        = 36;
  const STAT_EXPERTISE    = 37;
  const STAT_ATTACK_POWER = 38;
  const STAT_RANGED_ATTACK_POWER = 39;
  const STAT_FERAL_ATTACK_POWER  = 40;
  const STAT_SPELL_HEALING       = 41;
  const STAT_SPELL_DAMAGE        = 42;
  const STAT_MANA_REGENERATION   = 43;
  const STAT_ARMOR_PENETRATION   = 44;
  const STAT_SPELL_POWER         = 45;
  const STAT_HEALTH_REGENERATION = 46;
  const STAT_SPELL_PENETRATION   = 47;
  const STAT_BLOCK_VALUE         = 48;
  const STAT_MASTERY             = 49;
  const STAT_BONUS_ARMOR         = 50;
  const STAT_FIRE_RESISTANCE     = 51;
  const STAT_FROST_RESISTANCE    = 52;
  const STAT_SHADOW_RESISTANCE   = 54;
  const STAT_NATURE_RESISTANCE   = 55;
  const STAT_ARCANE_RESISTANCE   = 56;
  
  /**
   * Local copy of the specified item ID.
   *
   * @var int
   */
  protected $itemId;
  
  /**
   * Constructs a BlizzardApiWowItem object.
   * 
   * @param int $itemId
   *   The ID of the item to get.
   * @param array $options
   *  (optional) An array containing any combination of the following keys:
   *   - region: The Battle.net region to request data from.
   *   - language: The language to request data in.
   */
  public function __construct($itemId, $options = array()) {
    $options += array(
      'region' => self::REGION_US,
      'language' => self::LANGUAGE_DEFAULT
    );
    
    parent::__construct($options['region'], $options['language']);
    $this->setApiType('Item')
      ->setApiPath($this->apiPath . '/item/{itemid}')
      ->setItemId($itemId);
  }
  
  /**
   * Returns the item ID number.
   */
  public function getItemId() {
    return $this->itemId;
  }
  
  /**
   * Given a reforge ID, determine which stats were converted.
   *
   * @param int $reforgeId
   *   The stat's reforge ID. This number should be between 113 (spirit to
   *   dodge) and 168 (mastery to expertise).
   * @param bool $assoc
   *   (optional) If TRUE, return an associative array instead of an object.
   *   Defaults to FALSE.
   *
   * @return mixed
   *   An object (or associative array) containing the reforge information.
   *
   * @throws DomainException
   */
  public static function getReforgeStats($reforgeId, $assoc = FALSE) {
    if ($reforgeId < 113 || $reforgeId > 168) {
      throw new DomainException('Invalid reforge ID');
    }
    
    $table = self::getReforgeTable();
    $row = (int) (($reforgeId - 113) / 7);
    $col = $reforgeId - (is_null($table[$row][0]) ? 112 : $table[$row][0]);
    
    // You can't reforge to the same stat.
    if (is_null($table[$row][$col])) {
      $col++;
    }
    
    $obj = new stdClass();
    $obj->from = $row;
    $obj->to = $col;
    
    return $assoc ? (array) $obj : $obj;
  }
  
  /**
   * Returns a table of reforge options.
   *
   * @code
   *         Spirit  Dodge   Parry   Hit     Crit    Haste   Exp     Mastery
   * Spirit  x       113     114     115     116     117     118     119
   * Dodge   120     x       121     122     123     124     125     126
   * Parry   127     128     x       129     130     131     132     133
   * Hit     134     135     136     x       137     138     139     140
   * Crit    141     142     143     144     x       145     146     147
   * Haste   148     149     150     151     152     x       153     154
   * Exp     155     156     157     158     159     160     x       161
   * Mastery 162     163     164     165     166     167     168     x
   * @endcode
   *
   * @see BlizzardApiWowItem::getReforgeStats()
   */
  public static function getReforgeTable() {
    return array(
      //                               spi   dodge parry hit   crit  haste exp   mastery
      self::REFORGE_SPIRIT    => array(NULL, 113,  114,  115,  116,  117,  118,  119),
      self::REFORGE_DODGE     => array(120,  NULL, 121,  122,  123,  124,  125,  126),
      self::REFORGE_PARRY     => array(127,  128,  NULL, 129,  130,  131,  132,  133),
      self::REFORGE_HIT       => array(134,  135,  136,  NULL, 137,  138,  139,  140),
      self::REFORGE_CRIT      => array(141,  142,  143,  144,  NULL, 145,  146,  147),
      self::REFORGE_HASTE     => array(148,  149,  150,  151,  152,  NULL, 153,  154),
      self::REFORGE_EXPERTISE => array(155,  156,  157,  158,  159,  160,  NULL, 161),
      self::REFORGE_MASTERY   => array(162,  163,  164,  165,  166,  167,  168,  NULL),
    );
  }
  
  /**
   * Set the item to retrieve.
   *
   * @param int $itemId
   *   The ID of the item to get.
   *
   * @throws DomainException, InvalidArgumentException
   */
  public function setItemId($itemId) {
    if (!is_numeric($itemId)) {
      throw new InvalidArgumentException('Item ID must be a number');
    }
    if ($itemId < 1) {
      throw new DomainException('Item ID must be greater than 0');
    }
    
    $this->itemId = $itemId;
    $this->addPlaceholder('{itemid}', $itemId);
    return $this;
  }
}

/**
 * Provides information on an item set.
 *
 * Default cache lifetime: 30 days.
 *
 * @see BlizzardApiWowItem
 */
class BlizzardApiWowItemSet extends BlizzardApiWow {
  /**
   * Local copy of the item set ID.
   */
  protected $setId;
  
  /**
   * Constructs a BlizzardApiWowItemSet object.
   *
   * @param int $setId
   *   The ID number of the item set.
   * @param array $options
   *   (optional) An array containing any combination of the following keys:
   *   - region: The Battle.net region to request data from.
   *   - language: The language to request data in.
   */
  public function __construct($setId, $options = array()) {
    $options += array(
      'region' => self::REGION_US,
      'language' => self::LANGUAGE_DEFAULT
    );
    
    parent::__construct($options['region'], $options['language']);
    $this->setApiType('Item set')
      ->setApiPath($this->apiPath . '/item/set/{setid}')
      ->setItemSetId($setId);
  }
  
  /**
   * Returns the item set ID.
   */
  public function getItemSetId() {
    return $this->setId;
  }
  
  /**
   * Set the item set to retrieve.
   *
   * @param int $setId
   *   The ID number of the item set.
   * 
   * @throws InvalidArgumentException
   */
  public function setItemSetId($setId) {
    if (!is_numeric($setId)) {
      throw new InvalidArgumentException('Set ID must be a number');
    }
    
    $this->setId = $setId;
    $this->addPlaceholder('{setid}', $setId);
    return $this;
  }
}

/**
 * Provides quest information.
 *
 * Default cache lifetime: 30 days.
 */
class BlizzardApiWowQuest extends BlizzardApiWow {
  /**
   * Local copy of the specified quest ID.
   *
   * @var int
   */
  protected $questId;
  
  /**
   * Constructs a BlizzardApiQuest object.
   *
   * @param int $questId
   *   The ID number of the quest.
   * @param array $options
   *   (optional) An array containing any combination of the following keys:
   *   - region: The Battle.net region to request data from.
   *   - language: The language to request data in.
   */
  public function __construct($questId, $options = array()) {
    $options += array(
      'region' => self::REGION_US,
      'language' => self::LANGUAGE_DEFAULT
    );
    
    parent::__construct($options['region'], $options['language']);
    $this->setApiType('Quest')
     ->setApiPath($this->apiPath . '/quest/{questid}')
     ->setQuestId($questId);
  }
  
  /**
   * Returns the quest ID.
   */
  public function getQuestId() {
    return $this->questId;
  }
  
  /**
   * Set the quest to retrieve.
   *
   * @param int $questId
   *   The ID number of the quest.
   *
   * @throws DomainException, InvalidArgumentException
   */
  public function setQuestId($questId) {
    if (!is_numeric($questId)) {
      throw new InvalidArgumentException('Quest ID must be a number');
    }
    if ($questId < 1) {
      throw new DomainException('Quest ID must be greater than 0');
    }
    
    $this->questId = $questId;
    $this->addPlaceholder('{questid}', $questId);
    return $this;
  }
}

/**
 * Provides rated battleground ladder information for a Battle.net region.
 *
 * Developer note: Results from this API may still include invalid characters 
 * [@link http://us.battle.net/wow/en/forum/topic/4663258100#2 source @endlink].
 *
 * Default cache lifetime: 1 day.
 * 
 * @see BlizzardApiWowArenaLadder
 */
class BlizzardApiWowRatedBattlegroundLadder extends BlizzardApiWow {
  /**
   * Local copy of the ladder sorting method.
   *
   * @var bool
   */
  protected $sort;
  
  /**
   * Constructs a BlizzardApiWowRatedBattlegroundLadder object.
   *
   * @param int $size
   *   (optional) The number of entries per page. Defaults to 50.
   * @param int $page
   *   (optional) The page number to retrieve. Defaults to 1.
   * @param bool $asc
   *   (optional) If TRUE, sort the ladder in ascending order. Defaults to TRUE.
   * @param array $options
   *   (optional) An array containing any combination of the following keys:
   *   - region: The Battle.net region to request data from.
   *   - language: The language to request data in.
   */
  public function __construct($size = 50, $page = 1, $asc = TRUE, $options = array()) {
    $options += array(
      'region' => self::REGION_US,
      'language' => self::LANGUAGE_DEFAULT
    );
    
    parent::__construct($options['region'], $options['language']);
    $this->setApiType('Rated battleground')
      ->setApiPath($this->apiPath . '/pvp/ratedbg/ladder')
      ->setCacheLifetime(86400)  // 1 day
      ->setSize($size)->setPage($page)->setSortOrder($asc);
  }
  
  /**
   * Returns the page number to retrieve.
   *
   * @see BlizzardApiWowRatedBattlegroundLadder::getSize()
   */
  public function getPage() {
    return $this->fields['page'];
  }
  
  /**
   * Returns the number of entries allowed per page.
   *
   * @see BlizzardApiWowRatedBattlegroundLadder::getPage()
   */
  public function getSize() {
    return $this->fields['size'];
  }
  
  /**
   * Returns TRUE if the rated battleground ladder is sorted in ascending order,
   * FALSE otherwise.
   */
  public function getSortOrder() {
    return $this->sort;
  }
  
  /**
   * Overrides BlizzardApi::prepareData().
   */
  protected function prepareData() {
    $this->data = $this->data['bgRecord'];
  }
  
  /**
   * Set the rated battleground ladder page to retrieve.
   * 
   * @param int $page
   *   (optional) The ladder page to retrieve. Defaults to 1.
   *
   * @see BlizzardApiWowRatedBattlegroundLadder::setSize()
   *
   * @throws DomainException, InvalidArgumentException
   */
  public function setPage($page = 1) {
    if (!is_numeric($page)) {
      throw new InvalidArgumentException('Page must be a number');
    }
    if ($page < 1) {
      throw new DomainException('Page cannot be less than 1');
    }
    $this->fields['page'] = $page;
    return $this;
  }
  
  /**
   * Set the number of results to return per rated battleground ladder page.
   * 
   * @param int $size
   *   (optional) The number of entries per page. Defaults to 50.
   *
   * @see BlizzardApiWowRatedBattlegroundLadder::setPage()
   *
   * @throws DomainException, InvalidArgumentException
   */
  public function setSize($size = 50) {
    if (!is_numeric($size)) {
      throw new InvalidArgumentException('Size must be a number');
    }
    if ($size < 0) {
      throw new DomainException('Size cannot be less than 0');
    }
    $this->fields['size'] = $size;
    return $this;
  }
  
  /**
   * Set the order of the retrieved arena ladder.
   * 
   * @param bool $asc
   *   (optional) If TRUE, the ladder will be sorted in ascending order.
   *   Defaults to TRUE.
   */
  public function setSortOrder($asc = TRUE) {
    $this->sort = (bool) $asc;
    $this->fields['asc'] = $asc ? 'true' : 'false';
    return $this;
  }
}

/**
 * Provides realm status information.
 *
 * Default cache lifetime: 10 minutes.
 */
class BlizzardApiWowRealm extends BlizzardApiWow {
  const POPULATION_LOW    = 'low';
  const POPULATION_MEDIUM = 'medium';
  const POPULATION_HIGH   = 'high';
  
  const PVP_AREA_UNKNOWN    = -1;
  const PVP_AREA_IDLE       = 0;
  const PVP_AREA_POPULATING = 1;
  const PVP_AREA_ACTIVE     = 2;
  const PVP_AREA_CONCLUDED  = 3;
  
  const TYPE_PVE   = 'pve';
  const TYPE_PVP   = 'pvp';
  const TYPE_RP    = 'rp';
  const TYPE_RPPVP = 'rppvp';
  
  const STATUS_OFFLINE = 0;
  const STATUS_ONLINE  = 1;
  
  const QUEUE_NO  = 0;
  const QUEUE_YES = 1;
  
  private static $schema = array(
    'type' => array(
      self::TYPE_PVE,
      self::TYPE_PVP,
      self::TYPE_RP,
      self::TYPE_RPPVP
    ),
    'population' => array(
      self::POPULATION_LOW,
      self::POPULATION_MEDIUM,
      self::POPULATION_HIGH
    ),
    'status' => array(
      self::STATUS_OFFLINE,
      self::STATUS_ONLINE
    ),
    'queue' => array(
      self::QUEUE_NO,
      self::QUEUE_YES
    )
  );
  
  /**
   * Constructs a BlizzardApiWowRealm object.
   *
   * @param array $realms
   *   (optional) A list of realms to limit the results to. If not provided, the
   *   status of all realms will be returned.
   * @param array $options
   *   (optional) An array containing any combination of the following keys:
   *   - region: The Battle.net region to request data from.
   *   - language: The language to request data in.
   */
  public function __construct($realms = array(), $options = array()) {
    $options += array(
      'region' => self::REGION_US,
      'language' => self::LANGUAGE_DEFAULT
    );
    
    parent::__construct($options['region'], $options['language']);
    $this->setApiType('Realm status')
      ->setApiPath($this->apiPath . '/realm/status')
      ->setCacheLifetime(600)  // 10 minutes
      ->setRealms($realms);
  }
  
  public static function filterByPopulation($realms, $population) {
    if (!in_array($population, self::$schema['population'])) {
      throw new InvalidArgumentException("Invalid realm population: $population");
    }
    return parent::filterBy($realms, 'population', $population);
  }
  
  public static function filterByQueue($realms, $queue) {
    if (!in_array($queue, self::$schema['queue'])) {
      throw new InvalidArgumentException("Invalid realm queue: $queue");
    }
    return parent::filterBy($realms, 'queue', $queue);
  }
  
  public static function filterByStatus($realms, $status) {
    if (!in_array($status, self::$schema['status'])) {
      throw new InvalidArgumentException("Invalid realm status: $status");
    }
    return parent::filterBy($realms, 'status', $status);
  }
  
  public static function filterByType($realms, $type) {
    if (!in_array($type, self::$schema['type'])) {
      throw new InvalidArgumentException("Invalid realm type: $type");
    }
    return parent::filterBy($realms, 'type', $type);
  }
  
  //public static function filterByBattlegroup($realms, $battlegroup) {}
  
  /**
   * Overrides BlizzardApi::prepareData().
   */
  protected function prepareData() {
    $this->data = $this->data['realms'];
  }
  
  /**
   * Set the specific realms to get information on.
   *
   * @param array $realms
   *   (optional) A list of realms to limit the results to. If not provided, the
   *   status of all realms will be returned.
   */
  public function setRealms($realms = array()) {
    $this->fields['realms'] = $realms;
    return $this;
  }
}

/**
 * Provides recipe information.
 *
 * Default cache lifetime: 30 days.
 */
class BlizzardApiWowRecipe extends BlizzardApiWow {
  /**
   * Local copy of the specified recipe ID.
   * 
   * @var int
   */
  protected $recipeId;
  
  /**
   * Constructs a BlizzardApiWowRecipe object.
   *
   * @param int $recipeId
   *   The ID number of the recipe.
   * @param array $options
   *  (optional) An array containing any combination of the following keys:
   *   - region: The Battle.net region to request data from.
   *   - language: The language to request data in.
   */
  public function __construct($recipeId, $options = array()) {
    $options += array(
      'region' => self::REGION_US,
      'language' => self::LANGUAGE_DEFAULT
    );
    
    parent::__construct($options['region'], $options['language']);
    $this->setApiType('Recipe')
      ->setApiPath($this->apiPath . '/recipe/{recipeid}')
      ->setRecipeId($recipeId);
  }
  
  /**
   * Returns the recipe ID.
   */
  public function getRecipeId() {
    return $this->recipeId;
  }
  
  /**
   * Set the recipe to retrieve.
   *
   * @param int $recipeId
   *   The ID number of the recipe.
   * 
   * @throws DomainException, InvalidArgumentException
   */
  public function setRecipeId($recipeId) {
    if (!is_numeric($recipeId)) {
      throw new InvalidArgumentException('Recipe ID must be a number');
    }
    if ($recipeId < 1) {
      throw new DomainException('Recipe ID must be greater than 0');
    }
    
    $this->recipeId = $recipeId;
    $this->addPlaceholder('{recipeid}', $recipeId);
    return $this;
  }
}

/**
 * Provides information that compliments other resources and provides structure,
 * definition, and/or context.
 *
 * Default cache lifetime: 30 days.
 */
class BlizzardApiWowData extends BlizzardApiWow {
  /**
   * A list of battlegroups in the given region.
   */
  const RESOURCE_BATTLEGROUPS = 1;
  /**
   * A list of all achievements that a character can earn as well as the
   * category structure and hierarchy.
   */
  const RESOURCE_CHARACTER_ACHIEVEMENTS = 2;
  /**
   * A list of character classes.
   */
  const RESOURCE_CLASSES = 3;
  /**
   * A list of all achievements that guilds can earn as well as the category
   * structure and hierarchy.
   */
  const RESOURCE_GUILD_ACHIEVEMENTS = 4;
  /**
   * A list of guild perks.
   */
  const RESOURCE_GUILD_PERKS = 5;
  /**
   * A list of guild rewards.
   */
  const RESOURCE_GUILD_REWARDS = 6;
  /**
   * A list of item classes.
   */
  const RESOURCE_ITEM_CLASSES = 7;
  /**
   * A list of character races.
   */
  const RESOURCE_RACES = 8;
  
  /**
   * Each method in this class uses a unique API path, so we store a copy of the
   * base path during initialization.
   *
   * @var string
   */
  private $baseApiPath;
  
  /**
   * @var int
   */
  protected $dataResource;
  
  /**
   * Constructs a BlizzardApiWowData object.
   *
   * @param int $resource
   *   The type of data to get. See the resource constants provided by this
   *   class for supported values.
   * @param array $options
   *   (optional) An array containing any combination of the following keys:
   *   - region: The Battle.net region to request data from.
   *   - language: The language to request data in.
   * @param array $args
   *   (optional) Reserved for future use.
   */
  public function __construct($resource, $options = array(), $args = array()) {
    $options += array(
      'region' => self::REGION_US,
      'language' => self::LANGUAGE_DEFAULT
    );
    
    parent::__construct($options['region'], $options['language']);
    $this->baseApiPath = $this->apiPath;
    $this->setResourceType($resource, $args);
  }
  
  /**
   * Returns the type of data to request.
   */
  public function getResourceType() {
    return $this->dataResource;
  }
  
  /**
   * Overrides BlizzardApi::prepareData().
   */
  protected function prepareData() {
    switch ($this->dataResource) {
      case self::RESOURCE_BATTLEGROUPS:
        $this->data = $this->data['battlegroups'];
        break;
      case self::RESOURCE_CHARACTER_ACHIEVEMENTS:
      case self::RESOURCE_GUILD_ACHIEVEMENTS:
        $this->data = $this->data['achievements'];
        break;
      case self::RESOURCE_CLASSES:
      case self::RESOURCE_ITEM_CLASSES:
        $this->data = $this->data['classes'];
        break;
      case self::RESOURCE_GUILD_PERKS:
        $this->data = $this->data['perks'];
        break;
      case self::RESOURCE_GUILD_REWARDS:
        $this->data = $this->data['rewards'];
        break;
      case self::RESOURCE_RACES:
        $this->data = $this->data['races'];
        break;
    }
  }
  
  /**
   * Set the data resource type to request.
   *
   * @param int $resource
   *   The type of data to get. See the resource constants provided by this
   *   class for supported values.
   * @param array $args
   *   (optional) Reserved for future use.
   *
   * @throws InvalidArgumentException
   */
  public function setResourceType($resource, $args = array()) {
    switch ($resource) {
      case self::RESOURCE_BATTLEGROUPS:
        $this->setApiType('Battlegroups')
          ->setApiPath($this->baseApiPath . '/data/battlegroups/{battlegroup}')
          ->addPlaceholder('{battlegroup}', '');
        break;
      case self::RESOURCE_CHARACTER_ACHIEVEMENTS:
        $this->setApiType('Character achievements')
          ->setApiPath($this->baseApiPath . '/data/character/achievements');
        break;
      case self::RESOURCE_CLASSES:
        $this->setApiType('Classes')
          ->setApiPath($this->baseApiPath . '/data/character/classes');
        break;
      case self::RESOURCE_GUILD_ACHIEVEMENTS:
        $this->setApiType('Guild achievements')
          ->setApiPath($this->baseApiPath . '/data/guild/achievements');
        break;
      case self::RESOURCE_GUILD_PERKS:
        $this->setApiType('Guild perks')
          ->setApiPath($this->baseApiPath . '/data/guild/perks');
        break;
      case self::RESOURCE_GUILD_REWARDS:
        $this->setApiType('Guild rewards')
          ->setApiPath($this->baseApiPath . '/data/guild/rewards');
        break;
      case self::RESOURCE_ITEM_CLASSES:
        $this->setApiType('Item classes')
          ->setApiPath($this->baseApiPath . '/data/item/classes');
        break;
      case self::RESOURCE_RACES:
        $this->setApiType('Races')
          ->setApiPath($this->baseApiPath . '/data/character/races');
        break;
      default:
        throw new InvalidArgumentException("Invalid data resource type: $resource");
    }
    
    $this->dataResource = $resource;
    return $this;
  }
}

/**
 * Factory class to generate Blizzard API objects for World of Warcraft.
 */
class BlizzardApiWowResource {
  public static function getAchievement($achievementId, $options = array()) {
    return new BlizzardApiWowAchievement($achievementId, $options);
  }
  
  public static function getArenaLadder($battlegroup, $teamSize, $size = 50, $page = 1, $asc = TRUE, $options = array()) {
    return new BlizzardApiWowArenaLadder($battlegroup, $teamSize, $size, $page, $asc, $options);
  }
  
  public static function getArenaTeam($realm, $teamSize, $teamName, $options = array()) {
    return new BlizzardApiWowArenaTeam($realm, $teamSize, $teamName, $options);
  }
  
  public static function getAuction($realm, $options = array()) {
    return new BlizzardApiWowAuction($realm, $options);
  }
  
  public static function getCharacter($character, $realm, $fields = array(), $options = array()) {
    return new BlizzardApiWowCharacter($character, $realm, $fields, $options);
  }
  
  public static function getData($resource, $options = array(), $args = array()) {
    return new BlizzardApiWowData($resource, $options, $args);
  }
  
  public static function getGuild($guild, $realm, $fields = array(), $options = array()) {
    return new BlizzardApiWowGuild($guild, $realm, $fields, $options);
  }
  
  public static function getItem($itemId, $options = array()) {
    return new BlizzardApiWowItem($itemId, $options);
  }
  
  public static function getItemSet($setId, $options = array()) {
    return new BlizzardApiWowItemSet($setId, $options);
  }
  
  public static function getQuest($questId, $options = array()) {
    return new BlizzardApiWowQuest($questId, $options);
  }
  
  public static function getRatedBattlegroundLadder($size = 50, $page = 1, $asc = TRUE, $options = array()) {
    return new BlizzardApiWowRatedBattlegroundLadder($size, $page, $asc, $options);
  }
  
  public static function getRealmStatus($realms = array(), $options = array()) {
    return new BlizzardApiWowRealm($realms, $options);
  }
  
  public static function getRecipe($recipeId, $options = array()) {
    return new BlizzardApiWowRecipe($recipeId, $options);
  }
}
